# Example robodoc.rc
#
items:
    NAME
    SYNOPSIS
    PROGRAM
    FUNCTION
    SUBROUTINE
    MODULE
    INPUTS
    OUTPUTS
    USAGE
    PUBLIC
preformatted items:
ignore files:
options:
    --doc html
    --one_file_per_header 
    --src ..
    --html 
    --multidoc 
    --documenttitle "OpenCoarrays" 
    --index 
    --tabsize 8 
    --nopre 
    --sections
    --css robodoc.css
    --syntaxcolors
headertypes:
    p  "Programs"              robo_programs      1
    m  "Modules"               robo_modules       2
    d  "Derived types"         robo_derived_types 3
    t  "Type-bound procedures" robo_type_bound_procedures  3
    g  "Generic Interfaces"    robo_constructor_functions 3
    f  "Functions"             robo_functions   3
    s  "Subroutines"           robo_subroutines 3
    l  "Libraries"             robo_libraries 1
    e  "Tests"                 robo_tests 1
ignore files:
    README
    *~
accept files:
    *.f90
    *.F90
    *.c
    *.cpp
keywords:
    program 
    use 
    iso_fortran_env
    implicit 
    none
    integer
    parameter
    real
    kind 
    allocatable
    allocate
    do
    enddo
    end
    sync 
    all
    if
    deallocate
    num_images
    this_image
